"""A module to read Plink's binary files."""

# This file is part of pyplink.
#
# The MIT License (MIT)
#
# Copyright (c) 2014 Louis-Philippe Lemieux Perreault
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from .pyplink import PyPlink

try:
    from .version import pyplink_version as __version__
except ImportError:
    __version__ = None


__author__ = "Louis-Philippe Lemieux Perreault"
__copyright__ = "Copyright 2014 Louis-Philippe Lemieux Perreault"
__credits__ = ["Louis-Philippe Lemieux Perreault"]
__license__ = "MIT"
__maintainer__ = "Louis-Philippe Lemieux Perreault"
__email__ = "louis-philippe.lemieux.perreault@statgen.org"
__status__ = "Development"


def test(verbosity=1):  # pragma: no cover
    """Executes all the tests for pyplink.

    Args:
        verbosity (int): The verbosity level for :py:mod:`unittest`.

    Just set ``verbosity`` to an integer higher than ``1`` to have more
    information about the tests.

    """
    import unittest
    from .tests import test_suite

    # Testing
    unittest.TextTestRunner(verbosity=verbosity).run(test_suite)



"""Module that reads binary Plink files."""

# This file is part of pyplink.
#
# The MIT License (MIT)
#
# Copyright (c) 2014 Louis-Philippe Lemieux Perreault
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import os
import logging
from itertools import repeat
from collections import Counter
from io import UnsupportedOperation

try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest

import numpy as np
import pandas as pd

from six.moves import range


__author__ = "Louis-Philippe Lemieux Perreault"
__copyright__ = "Copyright 2014 Louis-Philippe Lemieux Perreault"
__license__ = "MIT"


__all__ = ["PyPlink"]


# The logger
logger = logging.getLogger(__name__)


# The recoding values
_geno_recode = {1: -1,  # Unknown genotype
                2:  1,  # Heterozygous genotype
                0:  2,  # Homozygous A1
                3:  0}  # Homozygous A2
_byte_recode = dict(value[::-1] for value in _geno_recode.items())


class PyPlink(object):
    """Reads and store a set of binary Plink files.

    Args:
        prefix (str): The prefix of the binary Plink files.
        mode (str): The open mode for the binary Plink file.
        bed_format (str): The type of bed (SNP-major or INDIVIDUAL-major).

    Reads or write binary Plink files (BED, BIM and FAM).

    .. code-block:: python

        from pyplink import PyPlink

        # Reading BED files
        with PyPlink("plink_file_prefix") as bed:
            pass

        # Writing BED files
        with PyPlink("plink_file_prefix", "w") as bed:
            pass

    """

    # The genotypes values
    _geno_values = np.array(
        [
            [_geno_recode[(i >> j) & 3] for j in range(0, 7, 2)]
            for i in range(256)
        ],
        dtype=np.int8,
    )

    def __init__(self, prefix, mode="r", bed_format="SNP-major"):
        """Initializes a new PyPlink instance."""
        # The mode
        self._mode = mode

        # The bed format
        if bed_format not in {"SNP-major", "INDIVIDUAL-major"}:
            raise ValueError("invalid bed format: {}".format(bed_format))
        self._bed_format = bed_format

        # These are the name of the files
        self.bed_filename = "{}.bed".format(prefix)
        self.bim_filename = "{}.bim".format(prefix)
        self.fam_filename = "{}.fam".format(prefix)

        if self._mode == "r":
            if self._bed_format != "SNP-major":
                raise ValueError("only SNP-major format is supported "
                                 "with mode 'r'")

            # Checking that all the files exists (otherwise, error...)
            for filename in (self.bed_filename, self.bim_filename,
                             self.fam_filename):
                if not os.path.isfile(filename):
                    raise IOError("No such file: '{}'".format(filename))

            # Setting BIM and FAM to None
            self._bim = None
            self._fam = None

            # Reading the input files
            self._read_bim()
            self._read_fam()
            self._read_bed()

            # Where we're at
            self._n = 0

        elif self._mode == "w":
            # The dummy number of samples and bytes
            self._nb_values = None

            # Opening the output BED file
            self._bed = open(self.bed_filename, "wb")
            self._write_bed_header()

        else:
            raise ValueError("invalid mode: '{}'".format(self._mode))

    def __repr__(self):
        """The representation of the PyPlink object."""
        if self._mode == "r":
            return "PyPlink({:,d} samples; {:,d} markers)".format(
                self.get_nb_samples(),
                self.get_nb_markers(),
            )

        return 'PyPlink(mode="w")'

    def __iter__(self):
        """The __iter__ function."""
        if self._mode != "r":
            raise UnsupportedOperation("not available in 'w' mode")

        return self

    def __next__(self):
        """The __next__ function."""
        return self.next()

    def __enter__(self):
        """Entering the context manager."""
        return self

    def __exit__(self, *args):
        """Exiting the context manager."""
        self.close()

    def close(self):
        """Closes the BED file."""
        # Closing the BED file
        self._bed.close()

    def next(self):
        """Returns the next marker.

        Returns:
            tuple: The marker name as a string and its genotypes as a
            :py:class:`numpy.ndarray`.

        """
        if self._mode != "r":
            raise UnsupportedOperation("not available in 'w' mode")

        self._n += 1
        if self._n > self._nb_markers:
            raise StopIteration()

        return self._bim.index[self._n - 1], self._read_current_marker()

    def _read_current_marker(self):
        """Reads the current marker and returns its genotypes."""
        return self._geno_values[
            np.frombuffer(self._bed.read(self._nb_bytes), dtype=np.uint8)
        ].flatten(order="C")[:self._nb_samples]

    def seek(self, n):
        """Gets to a certain marker position in the BED file.

        Args:
            n (int): The index of the marker to seek to.

        """
        if self._mode != "r":
            raise UnsupportedOperation("not available in 'w' mode")

        if 0 <= n < self._nb_markers:
            self._n = n
            self._bed.seek(self._get_seek_position(n))

        else:
            # Invalid seek value
            raise ValueError("invalid position in BED: {}".format(n))

    def _get_seek_position(self, n):
        """Gets the seek position in the file (including special bytes).

        Args:
            n (int): The index of the marker to seek to.

        """
        return 3 + self._nb_bytes * n

    def _read_bim(self):
        """Reads the BIM file."""
        # Reading the BIM file and setting the values
        bim = pd.read_csv(self.bim_filename, delim_whitespace=True,
                          names=["chrom", "snp", "cm", "pos", "a1", "a2"],
                          dtype=dict(snp=str, a1=str, a2=str))

        # Saving the index as integer
        bim["i"] = bim.index

        # Checking for duplicated markers
        try:
            bim = bim.set_index("snp", verify_integrity=True)
            self._has_duplicated = False

        except ValueError:
            # Setting this flag to true
            self._has_duplicated = True

            # Finding the duplicated markers
            duplicated = bim.snp.duplicated(keep=False)
            duplicated_markers = bim.loc[duplicated, "snp"]
            duplicated_marker_counts = duplicated_markers.value_counts()

            # The dictionary that will contain information about the duplicated
            # markers
            self._dup_markers = {
                m: [] for m in duplicated_marker_counts.index
            }

            # Logging a warning
            logger.warning("Duplicated markers found")
            for marker, count in duplicated_marker_counts.items():
                logger.warning("  - {}: {:,d} times".format(marker, count))
            logger.warning("Appending ':dupX' to the duplicated markers "
                           "according to their location in the BIM file")

            # Renaming the markers
            counter = Counter()
            for i, marker in duplicated_markers.items():
                counter[marker] += 1
                new_name = "{}:dup{}".format(marker, counter[marker])
                bim.loc[i, "snp"] = new_name

                # Updating the dictionary containing the duplicated markers
                self._dup_markers[marker].append(new_name)

            # Resetting the index
            bim = bim.set_index("snp", verify_integrity=True)

        # Encoding the allele
        #   - The original 0 is the actual 2 (a1/a1)
        #   - The original 2 is the actual 1 (a1/a2)
        #   - The original 3 is the actual 0 (a2/a2)
        #   - The original 1 is the actual -1 (no call)
        allele_encoding = np.array(
            [bim.a2 * 2, bim.a1 + bim.a2, bim.a1 * 2,
             list(repeat("00", bim.shape[0]))],
            dtype="U2",
        )
        self._allele_encoding = allele_encoding.T

        # Saving the data in the object
        self._bim = bim[["chrom", "pos", "cm", "a1", "a2", "i"]]
        self._nb_markers = self._bim.shape[0]

    def get_bim(self):
        """Returns the BIM file.

        Returns:
            pandas.DataFrame: The BIM file.

        """
        if self._mode != "r":
            raise UnsupportedOperation("not available in 'w' mode")

        return self._bim.drop("i", axis=1)

    def get_nb_markers(self):
        """Returns the number of markers.

        Returns:
            int: The number of markers in the dataset.

        """
        if self._mode != "r":
            raise UnsupportedOperation("not available in 'w' mode")

        return self._nb_markers

    def get_duplicated_markers(self):
        """Returns the duplicated markers, if any.

        Returns:
            dict: The set of duplicated marker (might be empty).

        """
        if self._has_duplicated:
            return self._dup_markers
        else:
            return {}

    def _read_fam(self):
        """Reads the FAM file."""
        # Reading the FAM file and setting the values
        fam = pd.read_csv(self.fam_filename, delim_whitespace=True,
                          names=["fid", "iid", "father", "mother", "gender",
                                 "status"],
                          dtype=dict(fid=str, iid=str, father=str, mother=str))

        # Saving the data in the object
        self._fam = fam
        self._nb_samples = self._fam.shape[0]

    def get_fam(self):
        """Returns the FAM file.

        Returns:
            pandas.DataFrame: The FAM file.

        """
        if self._mode != "r":
            raise UnsupportedOperation("not available in 'w' mode")

        return self._fam.copy()

    def get_nb_samples(self):
        """Returns the number of samples.

        Returns:
            int: The number of samples in the dataset.

        """
        if self._mode != "r":
            raise UnsupportedOperation("not available in 'w' mode")

        return self._nb_samples

    def _read_bed(self):
        """Reads the BED file."""
        # Checking if BIM and BAM files were both read
        if (self._bim is None) or (self._fam is None):
            raise RuntimeError("no BIM or FAM file were read")

        # The number of bytes per marker
        self._nb_bytes = int(np.ceil(self._nb_samples / 4.0))

        # Checking the file is valid by looking at the first 3 bytes and the
        # last entry (correct size)
        with open(self.bed_filename, "rb") as bed_file:
            # Checking that the first two bytes are OK
            if (ord(bed_file.read(1)) != 108) or (ord(bed_file.read(1)) != 27):
                raise ValueError("not a valid BED file: "
                                 "{}".format(self.bed_filename))

            # Checking that the format is SNP-major
            if ord(bed_file.read(1)) != 1:
                raise ValueError("not in SNP-major format (please recode): "
                                 "{}".format(self.bed_filename))

            # Checking the last entry (for BED corruption)
            seek_index = self._get_seek_position(self._bim.iloc[-1, :].i)
            bed_file.seek(seek_index)
            geno = self._geno_values[
                np.frombuffer(bed_file.read(self._nb_bytes), dtype=np.uint8)
            ].flatten(order="C")[:self._nb_samples]
            if geno.shape[0] != self._nb_samples:
                raise ValueError("invalid number of entries: corrupted BED?")

        # Opening the file for the rest of the operations (reading 3 bytes)
        self._bed = open(self.bed_filename, "rb")
        self._bed.read(3)

    def _write_bed_header(self):
        """Writes the BED first 3 bytes."""
        # Writing the first three bytes
        final_byte = 1 if self._bed_format == "SNP-major" else 0
        self._bed.write(bytearray((108, 27, final_byte)))

    def iter_geno(self):
        """Iterates over genotypes from the beginning of the BED file.

        Returns:
            tuple: The name of the marker as a string, and its genotypes as a
            :py:class:`numpy.ndarray` (additive format).

        """
        if self._mode != "r":
            raise UnsupportedOperation("not available in 'w' mode")

        # Seeking back at the beginning of the file
        self.seek(0)

        # Return itself (the generator)
        return self

    def iter_acgt_geno(self):
        """Iterates over genotypes (ACGT format).

        Returns:
            tuple: The name of the marker as a string, and its genotypes as a
            :py:class:`numpy.ndarray` (ACGT format).

        """
        # Need to iterate over itself, and modify the actual genotypes
        for i, (marker, geno) in enumerate(self.iter_geno()):
            yield marker, self._allele_encoding[i][geno]

    def iter_geno_marker(self, markers, return_index=False):
        """Iterates over genotypes for a list of markers.

        Args:
            markers (list): The list of markers to iterate onto.
            return_index (bool): Wether to return the marker's index or not.

        Returns:
            tuple: The name of the marker as a string, and its genotypes as a
            :py:class:`numpy.ndarray` (additive format).

        """
        if self._mode != "r":
            raise UnsupportedOperation("not available in 'w' mode")

        # If string, we change to list
        if isinstance(markers, str):
            markers = [markers]

        # Iterating over all markers
        if return_index:
            for marker in markers:
                geno, seek = self.get_geno_marker(marker, return_index=True)
                yield marker, geno, seek
        else:
            for marker in markers:
                yield marker, self.get_geno_marker(marker)

    def iter_acgt_geno_marker(self, markers):
        """Iterates over genotypes for a list of markers (ACGT format).

        Args:
            markers (list): The list of markers to iterate onto.

        Returns:
            tuple: The name of the marker as a string, and its genotypes as a
            :py:class:`numpy.ndarray` (ACGT format).

        """
        # We iterate over the markers
        for snp, geno, s in self.iter_geno_marker(markers, return_index=True):
            # Getting the SNP position and converting to ACGT
            yield snp, self._allele_encoding[s][geno]

    def get_geno_marker(self, marker, return_index=False):
        """Gets the genotypes for a given marker.

        Args:
            marker (str): The name of the marker.
            return_index (bool): Wether to return the marker's index or not.

        Returns:
            numpy.ndarray: The genotypes of the marker (additive format).

        """
        if self._mode != "r":
            raise UnsupportedOperation("not available in 'w' mode")

        # Check if the marker exists
        if marker not in self._bim.index:
            raise ValueError("{}: marker not in BIM".format(marker))

        # Seeking to the correct position
        seek_index = self._bim.loc[marker, "i"]
        self.seek(seek_index)

        if return_index:
            return self._read_current_marker(), seek_index
        return self._read_current_marker()

    def get_acgt_geno_marker(self, marker):
        """Gets the genotypes for a given marker (ACGT format).

        Args:
            marker (str): The name of the marker.

        Returns:
            numpy.ndarray: The genotypes of the marker (ACGT format).

        """
        # Getting the marker's genotypes
        geno, snp_position = self.get_geno_marker(marker, return_index=True)

        # Returning the ACGT's format of the genotypes
        return self._allele_encoding[snp_position][geno]

    def write_genotypes(self, genotypes):
        """Write genotypes to binary file.

        Args:
            genotypes (numpy.ndarray): The genotypes to write in the BED file.

        """
        if self._mode != "w":
            raise UnsupportedOperation("not available in 'r' mode")

        # Initializing the number of samples if required
        if self._nb_values is None:
            self._nb_values = len(genotypes)

        # Checking the expected number of samples
        if self._nb_values != len(genotypes):
            raise ValueError("{:,d} samples expected, got {:,d}".format(
                self._nb_values,
                len(genotypes),
            ))

        # Writing to file
        byte_array = [
            g[0] | (g[1] << 2) | (g[2] << 4) | (g[3] << 6) for g in
            self._grouper((_byte_recode[geno] for geno in genotypes), 4)
        ]
        self._bed.write(bytearray(byte_array))

    @staticmethod
    def _grouper(iterable, n, fillvalue=0):
        """Collect data into fixed-length chunks or blocks.

        Args:
            n (int): The size of the chunk.
            fillvalue (int): The fill value.

        Returns:
            iterator: An iterator over the chunks.

        """
        args = [iter(iterable)] * n
        return zip_longest(fillvalue=fillvalue, *args)



"""PyPlink test suite."""
# This file is part of pyplink.
#
# The MIT License (MIT)
#
# Copyright (c) 2014 Louis-Philippe Lemieux Perreault
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from __future__ import print_function

import os
import sys
import stat
import random
import shutil
import zipfile
import platform
import unittest
from tempfile import mkdtemp
from io import UnsupportedOperation
from distutils.spawn import find_executable
from subprocess import check_call, PIPE, CalledProcessError

try:
    from itertools import zip_longest as zip
except ImportError:
    from itertools import izip_longest as zip

try:
    from urllib.request import urlretrieve
except ImportError:
    from urllib import urlretrieve

try:
    from unittest import mock
except ImportError:
    import mock

from pkg_resources import resource_filename

import numpy as np
import pandas as pd

from six.moves import range

from .. import pyplink


def get_plink(tmp_dir):
    """Gets the Plink binary, if required."""
    # Checking if Plink is in the path
    plink_path = "plink"
    if platform.system() == "Windows":
        plink_path += ".exe"

    if find_executable(plink_path) is None:
        print("Downloading Plink", file=sys.stderr)

        # The url for each platform
        url = ("http://statgen.org/wp-content/uploads/Softwares/"
               "plink-1.0.7/{filename}")

        # Getting the name of the file
        filename = ""
        if platform.system() == "Windows":
            filename = "plink-1.07-dos.zip"
        elif platform.system() == "Darwin":
            filename = "plink-1.07-mac-intel.zip"
        elif platform.system() == "Linux":
            if platform.architecture()[0].startswith("32"):
                filename = "plink-1.07-i686.zip"
            elif platform.architecture()[0].startswith("64"):
                filename = "plink-1.07-x86_64.zip"
            else:
                return None, "System not compatible for Plink"
        else:
            return None, "System not compatible for Plink"

        # Downloading Plink
        zip_path = os.path.join(tmp_dir, filename)
        try:
            urlretrieve(
                url.format(filename=filename),
                zip_path,
            )
        except:
            return None, "Plink's URL is not available"

        # Unzipping Plink
        with zipfile.ZipFile(zip_path, "r") as z:
            z.extractall(tmp_dir)
        plink_path = os.path.join(tmp_dir, os.path.splitext(filename)[0],
                                  plink_path)
        if not os.path.isfile(plink_path):
            return None, "Cannot use Plink"

        # Making the script executable
        if platform.system() in {"Darwin", "Linux"}:
            os.chmod(plink_path, stat.S_IRWXU)

    # Testing Plink works
    try:
        check_call([
            plink_path,
            "--noweb",
            "--help",
            "--out", os.path.join(tmp_dir, "execution_test")
        ], stdout=PIPE, stderr=PIPE)
    except CalledProcessError:
        return None, "Plink cannot be properly used"
    except IOError:
        return None, "Plink was not properly installed"

    return plink_path, "OK"


class TestPyPlink(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Creating a temporary directory
        cls.tmp_dir = mkdtemp(prefix="pyplink_test_")

        # Getting the BED/BIM/FAM files
        cls.bed = resource_filename(
            __name__,
            os.path.join("data", "test_data.bed"),
        )
        cls.bim = resource_filename(
            __name__,
            os.path.join("data", "test_data.bim"),
        )
        cls.fam = resource_filename(
            __name__,
            os.path.join("data", "test_data.fam"),
        )

        # Getting the prefix of the files
        cls.prefix = os.path.splitext(cls.bed)[0]

        # The list of markers
        cls.markers = ["rs10399749", "rs2949420", "rs2949421", "rs2691310",
                       "rs4030303", "rs4030300", "rs3855952", "rs940550",
                       "rs13328714", "rs11490937"]

        # The genotypes
        cls.genotypes = [[0, 0, 1], [0, 1, 0], [-1, -1, -1], [-1, -1, 1],
                         [0, 0, 0], [0, 0, 0], [0, 1, 2], [0, 0, 0], [1, 0, 0],
                         [0, 1, 0]]
        cls.acgt_genotypes = [["CC", "CC", "GC"], ["TT", "CT", "TT"],
                              ["00", "00", "00"], ["00", "00", "AT"],
                              ["GG", "GG", "GG"], ["CC", "CC", "CC"],
                              ["AA", "GA", "GG"], ["TT", "TT", "TT"],
                              ["GC", "CC", "CC"], ["GG", "AG", "GG"]]

        # Getting Plink
        cls.plink_path, cls.plink_message = get_plink(cls.tmp_dir)

    def setUp(self):
        # Reading the plink binary file
        self.pedfile = pyplink.PyPlink(self.prefix)

    @classmethod
    def tearDownClass(cls):
        # Cleaning the temporary directory
        shutil.rmtree(cls.tmp_dir)

    def tearDown(self):
        # Closing the PyPlink object
        self.pedfile.close()

    def test_pyplink_object_integrity(self):
        """Checks the integrity of the PyPlink object."""
        # Checking the name of the BED file
        self.assertTrue(hasattr(self.pedfile, "bed_filename"))
        self.assertEqual(self.bed, self.pedfile.bed_filename)

        # Checking the name of the BIM file
        self.assertTrue(hasattr(self.pedfile, "bim_filename"))
        self.assertEqual(self.bim, self.pedfile.bim_filename)

        # Checking the BIM object
        self.assertTrue(hasattr(self.pedfile, "_bim"))
        self.assertTrue(isinstance(self.pedfile._bim, pd.DataFrame))

        # Checking the name of the FAM file
        self.assertTrue(hasattr(self.pedfile, "fam_filename"))
        self.assertEqual(self.fam, self.pedfile.fam_filename)

        # Checking the FAM object
        self.assertTrue(hasattr(self.pedfile, "_fam"))
        self.assertTrue(isinstance(self.pedfile._fam, pd.DataFrame))

    def test_pyplink_object_error(self):
        """Checks what happens when we play with the PyPlink object."""
        # Changing the BIM to None
        ori = self.pedfile._bim
        self.pedfile._bim = None
        with self.assertRaises(RuntimeError) as cm:
            self.pedfile._read_bed()
        self.assertEqual("no BIM or FAM file were read", str(cm.exception))
        self.pedfile._bim = ori

        # Changing the FAM to None
        ori = self.pedfile._fam
        self.pedfile._fam = None
        with self.assertRaises(RuntimeError) as cm:
            self.pedfile._read_bed()
        self.assertEqual("no BIM or FAM file were read", str(cm.exception))
        self.pedfile._fam = ori

    def test_pyplink_bad_bed(self):
        """Checks what happens when we read a bad BED file."""
        # The new file prefix
        new_prefix = os.path.join(self.tmp_dir, "bad_data")

        # Copying the FAM file
        new_fam = new_prefix + ".fam"
        with open(new_fam, "w") as o_file, open(self.fam, "r") as i_file:
            o_file.write(i_file.read())

        # Copying the BIM file
        new_bim = new_prefix + ".bim"
        with open(new_bim, "w") as o_file, open(self.bim, "r") as i_file:
            o_file.write(i_file.read())

        # Creating a new BED file with invalid number of bytes
        new_bed = new_prefix + ".bed"
        with open(new_bed, "wb") as o_file:
            o_file.write(bytearray([108, 27, 1, 1, 2, 3, 4]))

        # This should raise an exception
        with self.assertRaises(ValueError) as cm:
            pyplink.PyPlink(new_prefix)
        self.assertEqual("invalid number of entries: corrupted BED?",
                         str(cm.exception))

        # Creating a new BED file with invalid first byte
        new_bed = new_prefix + ".bed"
        with open(new_bed, "wb") as o_file:
            o_file.write(bytearray([107, 27, 1, 1, 2, 3, 4]))

        # This should raise an exception
        with self.assertRaises(ValueError) as cm:
            pyplink.PyPlink(new_prefix)
        self.assertEqual("not a valid BED file: {}".format(new_bed),
                         str(cm.exception))

        # Creating a new BED file with invalid second byte
        new_bed = new_prefix + ".bed"
        with open(new_bed, "wb") as o_file:
            o_file.write(bytearray([108, 28, 1, 1, 2, 3, 4]))

        # This should raise an exception
        with self.assertRaises(ValueError) as cm:
            pyplink.PyPlink(new_prefix)
        self.assertEqual("not a valid BED file: {}".format(new_bed),
                         str(cm.exception))

        # Creating a new BED file not in SNP-major format
        new_bed = new_prefix + ".bed"
        with open(new_bed, "wb") as o_file:
            o_file.write(bytearray([108, 27, 0, 1, 2, 3, 4]))

        # This should raise an exception
        with self.assertRaises(ValueError) as cm:
            pyplink.PyPlink(new_prefix)
        self.assertEqual(
            "not in SNP-major format (please recode): {}".format(new_bed),
            str(cm.exception),
        )

    def test_missing_files(self):
        """Checks that an exception is raised when an input file is missing."""
        # Creating dummy BED/BIM/FAM files
        prefix = os.path.join(self.tmp_dir, "test_missing")
        for extension in (".bed", ".bim", ".fam"):
            with open(prefix + extension, "w"):
                pass

        # Removing the files (one by one) and checking the exception is raised
        for extension in (".bed", ".bim", ".fam"):
            os.remove(prefix + extension)
            with self.assertRaises(IOError) as cm:
                pyplink.PyPlink(prefix)
            self.assertEqual("No such file: '{}'".format(prefix + extension),
                             str(cm.exception))
            with open(prefix + extension, "w"):
                pass

    def test_get_nb_markers(self):
        """Tests that the correct number of markers is returned."""
        self.assertEqual(self.pedfile.get_nb_markers(), 10)

    def test_get_nb_markers_w_mode(self):
        """Tests that an exception is raised if in write mode."""
        with self.assertRaises(UnsupportedOperation) as cm:
            # Creating the dummy PyPlink object
            prefix = os.path.join(self.tmp_dir, "test_error")
            with pyplink.PyPlink(prefix, "w") as p:
                p.get_nb_markers()
        self.assertEqual("not available in 'w' mode", str(cm.exception))

    def test_get_nb_samples(self):
        """Tests that the correct number of samples is returned."""
        self.assertEqual(self.pedfile.get_nb_samples(), 3)

    def test_get_nb_samples_w_mode(self):
        """Tests that an exception is raised if in write mode."""
        with self.assertRaises(UnsupportedOperation) as cm:
            # Creating the dummy PyPlink object
            prefix = os.path.join(self.tmp_dir, "test_error")
            with pyplink.PyPlink(prefix, "w") as p:
                p.get_nb_samples()
        self.assertEqual("not available in 'w' mode", str(cm.exception))

    def test_get_bim(self):
        """Tests the 'get_bim' function."""
        # The original BIM file (with the 'i' column)
        ori_bim = self.pedfile._bim

        # The expected values
        chromosomes = [1, 2, 3, 4, 4, 5, 6, 6, 6, 8]
        positions = [45162, 45257, 45413, 46844, 72434, 72515, 77689, 78032,
                     81468, 222077]
        cms = [0, 1, 1, 2, 2, 3, 4, 4, 5, 6]
        a1s = ["G", "C", "0", "A", "0", "0", "G", "0", "G", "A"]
        a2s = ["C", "T", "0", "T", "G", "C", "A", "T", "C", "G"]

        # Getting the BIM file
        bim = self.pedfile.get_bim()

        # Checking the columns
        self.assertTrue(
            set(bim.columns.values) == {"chrom", "pos", "cm", "a1", "a2"}
        )

        # Checking the indexes
        self.assertTrue(set(bim.index.values) == set(self.markers))

        # Checking the values for the markers
        zipped = zip(self.markers, chromosomes, positions, cms, a1s, a2s)
        for marker, chrom, pos, cm, a1, a2 in zipped:
            self.assertEqual(chrom, bim.loc[marker, "chrom"])
            self.assertEqual(pos, bim.loc[marker, "pos"])
            self.assertEqual(cm, bim.loc[marker, "cm"])
            self.assertEqual(a1, bim.loc[marker, "a1"])
            self.assertEqual(a2, bim.loc[marker, "a2"])

        # Comparing with the original values
        comparison = ori_bim.loc[:, ["chrom", "pos", "cm", "a1", "a2"]] == bim
        self.assertTrue(comparison.all().all())

        # Testing that changing a values in the BIM, doesn't change the value
        # in the original BIM
        bim.loc["rs4030300", "chrom"] = 2
        bim.loc["rs2949420", "cm"] = 0.1
        comparison = ori_bim.loc[:, ["chrom", "pos", "cm", "a1", "a2"]] == bim
        self.assertFalse(comparison.all().chrom)
        self.assertFalse(comparison.all().cm)
        self.assertTrue(comparison.all()[["pos", "a1", "a2"]].all())

    def test_get_bim_w_mode(self):
        """Tests that an exception is raised if in write mode."""
        with self.assertRaises(UnsupportedOperation) as cm:
            # Creating the dummy PyPlink object
            prefix = os.path.join(self.tmp_dir, "test_error")
            with pyplink.PyPlink(prefix, "w") as p:
                p.get_bim()
        self.assertEqual("not available in 'w' mode", str(cm.exception))

    def test_get_fam(self):
        """Tests the 'get_fam' function."""
        # The original FAM file (with the 'byte' and 'bit' columns)
        ori_fam = self.pedfile._fam

        # The expected values
        fids = ["Sample_1", "Sample_2", "Sample_3"]
        iids = ["Sample_1", "Sample_2", "Sample_3"]
        fathers = ["0", "0", "Sample_1"]
        mothers = ["0", "0", "Sample_2"]
        genders = [1, 2, 2]
        status = [-9, -9, -9]

        # Getting the FAM file
        fam = self.pedfile.get_fam()

        # Checking the columns
        self.assertTrue(
            set(fam.columns.values) == {"fid", "iid", "father", "mother",
                                        "gender", "status"}
        )

        # Checking the values
        zipped = zip(fids, iids, fathers, mothers, genders, status)
        for i, (fid, iid, father, mother, gender, s) in enumerate(zipped):
            self.assertEqual(fid, fam.loc[i, "fid"])
            self.assertEqual(iid, fam.loc[i, "iid"])
            self.assertEqual(father, fam.loc[i, "father"])
            self.assertEqual(mother, fam.loc[i, "mother"])
            self.assertEqual(gender, fam.loc[i, "gender"])
            self.assertEqual(s, fam.loc[i, "status"])

        # Comparing with the original values
        comparison = ori_fam.loc[:, ["fid", "iid", "father", "mother",
                                     "gender", "status"]] == fam
        self.assertTrue(comparison.all().all())

        # Testing that changing a values in the FAM, doesn't change the value
        # in the original FAM
        fam.loc[2, "father"] = "0"
        fam.loc[0, "status"] = 2
        comparison = ori_fam.loc[:, ["fid", "iid", "father", "mother",
                                     "gender", "status"]] == fam
        self.assertFalse(comparison.all().father)
        self.assertFalse(comparison.all().status)
        self.assertTrue(
            comparison.all()[["fid", "iid", "mother", "gender"]].all()
        )

    def test_get_fam_w_mode(self):
        """Tests that an exception is raised if in write mode."""
        with self.assertRaises(UnsupportedOperation) as cm:
            # Creating the dummy PyPlink object
            prefix = os.path.join(self.tmp_dir, "test_error")
            with pyplink.PyPlink(prefix, "w") as p:
                p.get_fam()
        self.assertEqual("not available in 'w' mode", str(cm.exception))

    def test_generator(self):
        """Testing the class as a generator."""
        # Zipping and checking
        zipped = zip(
            [i for i in zip(self.markers, self.genotypes)],
            self.pedfile,
        )
        for (e_marker, e_geno), (marker, geno) in zipped:
            self.assertEqual(e_marker, marker)
            np.testing.assert_array_equal(e_geno, geno)

        # The generator should be empty
        remaining = [(marker, geno) for marker, geno in self.pedfile]
        self.assertEqual(0, len(remaining))

    def test_generator_w_mode(self):
        """Tests that an exception is raised if in write mode."""
        with self.assertRaises(UnsupportedOperation) as cm:
            # Creating the dummy PyPlink object
            prefix = os.path.join(self.tmp_dir, "test_error")
            with pyplink.PyPlink(prefix, "w") as p:
                marker, genotypes = next(p)
        self.assertEqual("not available in 'w' mode", str(cm.exception))

    def test_next(self):
        """Tests that an exception is raised when calling next in w mode."""
        marker, genotypes = self.pedfile.next()

        # Comparing
        self.assertEqual(self.markers[0], marker)
        np.testing.assert_array_equal(self.genotypes[0], genotypes)

    def test_next_w_mode(self):
        """Tests that an exception is raised when calling next in w mode."""
        with self.assertRaises(UnsupportedOperation) as cm:
            # Creating the dummy PyPlink object
            prefix = os.path.join(self.tmp_dir, "test_error")
            with pyplink.PyPlink(prefix, "w") as p:
                p.next()
        self.assertEqual("not available in 'w' mode", str(cm.exception))

    def test_seek(self):
        """Testing the seeking (for the generator)."""
        for marker, geno in self.pedfile:
            pass

        # The generator should be empty
        remaining = [(marker, geno) for marker, geno in self.pedfile]
        self.assertEqual(0, len(remaining))

        # Seeking at the second position
        zipped = zip(
            [i for i in zip(self.markers[1:], self.genotypes[1:])],
            self.pedfile,
        )
        self.pedfile.seek(1)
        for (e_marker, e_geno), (marker, geno) in zipped:
            self.assertEqual(e_marker, marker)
            np.testing.assert_array_equal(e_geno, geno)

        # Seeking at the fourth position
        zipped = zip(
            [i for i in zip(self.markers[3:], self.genotypes[3:])],
            self.pedfile,
        )
        self.pedfile.seek(3)
        for (e_marker, e_geno), (marker, geno) in zipped:
            self.assertEqual(e_marker, marker)
            np.testing.assert_array_equal(e_geno, geno)

        # Seeking at the tenth position
        zipped = zip(
            [i for i in zip(self.markers[9:], self.genotypes[9:])],
            self.pedfile,
        )
        self.pedfile.seek(9)
        for (e_marker, e_geno), (marker, geno) in zipped:
            self.assertEqual(e_marker, marker)
            np.testing.assert_array_equal(e_geno, geno)

        # Seeking at an invalid position
        with self.assertRaises(ValueError) as cm:
            self.pedfile.seek(-1)
        self.assertEqual("invalid position in BED: -1", str(cm.exception))

        # Seeking at an invalid position
        with self.assertRaises(ValueError) as cm:
            self.pedfile.seek(100)
        self.assertEqual("invalid position in BED: 100", str(cm.exception))

        # Seeking at an invalid position
        with self.assertRaises(ValueError) as cm:
            self.pedfile.seek(10)
        self.assertEqual("invalid position in BED: 10", str(cm.exception))

    def test_seek_w_mode(self):
        """Tests that an exception is raised if in write mode."""
        with self.assertRaises(UnsupportedOperation) as cm:
            # Creating the dummy PyPlink object
            prefix = os.path.join(self.tmp_dir, "test_error")
            with pyplink.PyPlink(prefix, "w") as p:
                p.seek(100)
        self.assertEqual("not available in 'w' mode", str(cm.exception))

    def test_iter_geno(self):
        """Tests the 'iter_geno' function."""
        zipped = zip(
            [i for i in zip(self.markers, self.genotypes)],
            self.pedfile.iter_geno(),
        )
        for (e_marker, e_geno), (marker, geno) in zipped:
            self.assertEqual(e_marker, marker)
            np.testing.assert_array_equal(e_geno, geno)

    def test_iter_geno_w_mode(self):
        """Tests that an exception is raised if in write mode."""
        with self.assertRaises(UnsupportedOperation) as cm:
            # Creating the dummy PyPlink object
            prefix = os.path.join(self.tmp_dir, "test_error")
            with pyplink.PyPlink(prefix, "w") as p:
                marker, genotypes = next(p.iter_geno())
        self.assertEqual("not available in 'w' mode", str(cm.exception))

    def test_iter_acgt_geno(self):
        """Tests the 'iter_acgt_geno" function."""
        zipped = zip(
            [i for i in zip(self.markers, self.acgt_genotypes)],
            self.pedfile.iter_acgt_geno(),
        )
        for (e_marker, e_geno), (marker, geno) in zipped:
            self.assertEqual(e_marker, marker)
            np.testing.assert_array_equal(e_geno, geno)

    def test_iter_acgt_geno_w_mode(self):
        """Tests that an exception is raised if in write mode."""
        with self.assertRaises(UnsupportedOperation) as cm:
            # Creating the dummy PyPlink object
            prefix = os.path.join(self.tmp_dir, "test_error")
            with pyplink.PyPlink(prefix, "w") as p:
                marker, genotypes = next(p.iter_acgt_geno())
        self.assertEqual("not available in 'w' mode", str(cm.exception))

    def test_iter_geno_marker(self):
        """Tests the 'iter_geno_marker' function."""
        # Getting a subset of indexes
        indexes = random.sample(range(len(self.markers)), 4)

        # Getting the markers and genotypes
        markers = [self.markers[i] for i in indexes]
        genotypes = [self.genotypes[i] for i in indexes]

        # Zipping and comparing
        zipped = zip(
            [i for i in zip(markers, genotypes)],
            self.pedfile.iter_geno_marker(markers),
        )
        for (e_marker, e_geno), (marker, geno) in zipped:
            self.assertEqual(e_marker, marker)
            np.testing.assert_array_equal(e_geno, geno)

        # Testing a single marker
        index = random.randint(0, len(self.markers) - 1)
        e_marker = self.markers[index]
        e_geno = self.genotypes[index]
        for marker, geno in self.pedfile.iter_geno_marker(e_marker):
            self.assertEqual(e_marker, marker)
            np.testing.assert_array_equal(e_geno, geno)

        # Adding a marker that doesn't exist
        markers.extend(["unknown_1", "unknown_2"])
        with self.assertRaises(ValueError) as cm:
            [i for i in self.pedfile.iter_geno_marker(markers)]
        self.assertEqual("unknown_1: marker not in BIM", str(cm.exception))

    def test_iter_geno_marker_w_mode(self):
        """Tests that an exception is raised if in write mode."""
        with self.assertRaises(UnsupportedOperation) as cm:
            # Creating the dummy PyPlink object
            prefix = os.path.join(self.tmp_dir, "test_error")
            with pyplink.PyPlink(prefix, "w") as p:
                marker, genotypes = next(p.iter_geno_marker(["M1", "M2"]))
        self.assertEqual("not available in 'w' mode", str(cm.exception))

    def test_iter_acgt_geno_marker(self):
        """Tests the 'iter_acgt_geno_marker' function."""
        # Getting a subset of indexes
        indexes = random.sample(range(len(self.markers)), 4)

        # Getting the markers and genotypes
        markers = [self.markers[i] for i in indexes]
        genotypes = [self.acgt_genotypes[i] for i in indexes]

        # Zipping and comparing
        zipped = zip(
            [i for i in zip(markers, genotypes)],
            self.pedfile.iter_acgt_geno_marker(markers),
        )
        for (e_marker, e_geno), (marker, geno) in zipped:
            self.assertEqual(e_marker, marker)
            np.testing.assert_array_equal(e_geno, geno)

        # Testing a single marker
        index = random.randint(0, len(self.markers) - 1)
        e_marker = self.markers[index]
        e_geno = self.acgt_genotypes[index]
        for marker, geno in self.pedfile.iter_acgt_geno_marker(e_marker):
            self.assertEqual(e_marker, marker)
            np.testing.assert_array_equal(e_geno, geno)

        # Adding a marker that doesn't exist
        markers.extend(["unknown_3", "unknown_4"])
        with self.assertRaises(ValueError) as cm:
            [i for i in self.pedfile.iter_acgt_geno_marker(markers)]
        self.assertEqual("unknown_3: marker not in BIM", str(cm.exception))

    def test_iter_acgt_geno_marker_w_mode(self):
        """Tests that an exception is raised if in write mode."""
        with self.assertRaises(UnsupportedOperation) as cm:
            # Creating the dummy PyPlink object
            prefix = os.path.join(self.tmp_dir, "test_error")
            with pyplink.PyPlink(prefix, "w") as p:
                marker, genotypes = next(p.iter_acgt_geno_marker(["M1", "M2"]))
        self.assertEqual("not available in 'w' mode", str(cm.exception))

    def test_repr_r_mode(self):
        """Tests the object representation of the string (r mode)."""
        # Counting the number of samples
        nb_samples = None
        with open(self.fam, "r") as i_file:
            nb_samples = len(i_file.read().splitlines())

        # Counting the number of markers
        nb_markers = None
        with open(self.bim, "r") as i_file:
            nb_markers = len(i_file.read().splitlines())

        # Creating the expected string representation
        e_repr = "PyPlink({:,d} samples; {:,d} markers)".format(nb_samples,
                                                                nb_markers)

        # Getting the observed string representation
        o_repr = str(self.pedfile)

        # Comparing
        self.assertEqual(e_repr, o_repr)

    def test_repr_w_mode(self):
        """Tests the object representation of the string (w mode)."""
        # The expected representation
        e_repr = 'PyPlink(mode="w")'

        # Creating the dummy PyPlink object
        prefix = os.path.join(self.tmp_dir, "test_repr")
        with pyplink.PyPlink(prefix, "w") as pedfile:
            # Comparing the expected with the observed representation
            o_repr = str(pedfile)
            self.assertEqual(e_repr, o_repr)

    def test_get_geno_marker(self):
        """Tests the 'get_geno_marker' function."""
        # Getting a random marker to test
        i = random.choice(range(len(self.markers)))
        marker = self.markers[i]
        e_geno = self.genotypes[i]

        # Getting the genotype
        o_geno = self.pedfile.get_geno_marker(marker)
        np.testing.assert_array_equal(o_geno, e_geno)

        # Asking for an unknown marker should raise an ValueError
        with self.assertRaises(ValueError) as cm:
            self.pedfile.get_geno_marker("dummy_marker")
        self.assertEqual(
            "dummy_marker: marker not in BIM",
            str(cm.exception),
        )

    def test_get_geno_marker_w_mode(self):
        """Tests that an exception is raised if in write mode."""
        with self.assertRaises(UnsupportedOperation) as cm:
            # Creating the dummy PyPlink object
            prefix = os.path.join(self.tmp_dir, "test_error")
            with pyplink.PyPlink(prefix, "w") as p:
                p.get_geno_marker("M1")
        self.assertEqual("not available in 'w' mode", str(cm.exception))

    def test_get_iter_w_mode(self):
        """Tests that an exception is raised if in write mode."""
        with self.assertRaises(UnsupportedOperation) as cm:
            # Creating the dummy PyPlink object
            prefix = os.path.join(self.tmp_dir, "test_error")
            with pyplink.PyPlink(prefix, "w") as p:
                iter(p)
        self.assertEqual("not available in 'w' mode", str(cm.exception))

    def test_get_acgt_geno_marker(self):
        """Tests the 'get_acgt_geno_marker' function."""
        # Getting a random marker to test
        i = random.choice(range(len(self.markers)))
        marker = self.markers[i]
        e_geno = self.acgt_genotypes[i]

        # Getting the genotype
        o_geno = self.pedfile.get_acgt_geno_marker(marker)
        np.testing.assert_array_equal(o_geno, e_geno)

        # Asking for an unknown marker should raise an ValueError
        with self.assertRaises(ValueError) as cm:
            self.pedfile.get_acgt_geno_marker("dummy_marker")
        self.assertEqual("dummy_marker: marker not in BIM", str(cm.exception))

    def test_get_acgt_geno_marker_w_mode(self):
        """Tests that an exception is raised if in write mode."""
        with self.assertRaises(UnsupportedOperation) as cm:
            # Creating the dummy PyPlink object
            prefix = os.path.join(self.tmp_dir, "test_error")
            with pyplink.PyPlink(prefix, "w") as p:
                p.get_acgt_geno_marker("M1")
        self.assertEqual("not available in 'w' mode", str(cm.exception))

    def test_get_context_read_mode(self):
        """Tests the PyPlink object as context manager."""
        with pyplink.PyPlink(self.prefix) as genotypes:
            self.assertEqual(3, len(genotypes.get_fam().head(n=3)))

    def test_invalid_mode(self):
        """Tests invalid mode when PyPlink as context manager."""
        with self.assertRaises(ValueError) as cm:
            pyplink.PyPlink(self.prefix, "u")
        self.assertEqual("invalid mode: 'u'", str(cm.exception))

    def test_write_binary(self):
        """Tests writing a Plink binary file."""
        # The expected genotypes
        expected_genotypes = [
            np.array([0,  0,  0, 1,  0, 1, 2], dtype=int),
            np.array([0,  0,  0, 0, -1, 0, 1], dtype=int),
            np.array([0, -1, -1, 2,  0, 0, 0], dtype=int),
        ]

        # The prefix
        test_prefix = os.path.join(self.tmp_dir, "test_write")

        # Writing the binary file
        with pyplink.PyPlink(test_prefix, "w") as pedfile:
            for genotypes in expected_genotypes:
                pedfile.write_genotypes(genotypes)

        # Checking the file exists
        self.assertTrue(os.path.isfile(test_prefix + ".bed"))

        # Writing the FAM file
        with open(test_prefix + ".fam", "w") as o_file:
            for i in range(7):
                print("f{}".format(i+1), "s{}".format(i+1), "0", "0",
                      random.choice((1, 2)), "-9", sep=" ", file=o_file)

        # Writing the BIM file
        with open(test_prefix + ".bim", "w") as o_file:
            for i in range(3):
                print(i+1, "m{}".format(i+1), "0", i+1, "T", "A", sep="\t",
                      file=o_file)

        # Reading the written binary file
        with pyplink.PyPlink(test_prefix) as pedfile:
            for i, (marker, genotypes) in enumerate(pedfile):
                self.assertEqual("m{}".format(i+1), marker)
                np.testing.assert_array_equal(expected_genotypes[i], genotypes)

    def test_write_binary_error(self):
        """Tests writing a binary file, with different number of sample."""
        # The expected genotypes
        expected_genotypes = [
            np.array([0, 0, 0, 1, 0, 1, 2], dtype=int),
            np.array([0, 0, 0, 0, -1, 0], dtype=int),
            np.array([0, -1, -1, 2, 0, 0, 0], dtype=int),
        ]

        # The prefix
        test_prefix = os.path.join(self.tmp_dir, "test_write")

        # Writing the binary file
        with self.assertRaises(ValueError) as cm:
            with pyplink.PyPlink(test_prefix, "w") as pedfile:
                pedfile.write_genotypes(expected_genotypes[0])  # 7 genotypes
                pedfile.write_genotypes(expected_genotypes[1])  # 6 genotypes
        self.assertEqual("7 samples expected, got 6", str(cm.exception))

    def test_grouper_padding(self):
        """Tests the _grouper function (when padding is required)."""
        expected_chunks = [
            (0, 1, 2),
            (3, 4, 5),
            (6, 7, 8),
            (9, 0, 0),
        ]
        observed_chunks = pyplink.PyPlink._grouper(range(10), 3)
        for expected, observed in zip(expected_chunks, observed_chunks):
            self.assertEqual(expected, observed)

    def test_grouper_no_padding(self):
        """Tests the _grouper function (when padding is not required)."""
        expected_chunks = [
            (0, 1, 2, 3, 4),
            (5, 6, 7, 8, 9),
        ]
        observed_chunks = pyplink.PyPlink._grouper(range(10), 5)
        for expected, observed in zip(expected_chunks, observed_chunks):
            self.assertEqual(expected, observed)

    @unittest.skipIf(platform.system() not in {"Darwin", "Linux", "Windows"},
                     "Plink not available for {}".format(platform.system()))
    def test_with_plink(self):
        """Tests to read a binary file using Plink."""
        # Checking if we need to skip
        if self.plink_path is None:
            self.skipTest(self.plink_message)

        # Creating the BED file
        all_genotypes = [
            [0, 1, 0, 0, -1, 0, 1, 0, 0, 2],
            [2, 1, 2, 2,  2, 2, 2, 1, 0, 1],
            [0, 0, 0, 0,  0, 1, 0, 0, 0, 0],
        ]
        prefix = os.path.join(self.tmp_dir, "test_output")
        with pyplink.PyPlink(prefix, "w") as pedfile:
            for genotypes in all_genotypes:
                pedfile.write_genotypes(genotypes)

        # Creating the FAM file
        fam_content = [
            ["F0", "S0", "0", "0", "1", "-9"],
            ["F1", "S1", "0", "0", "1", "-9"],
            ["F2", "S2", "0", "0", "2", "-9"],
            ["F3", "S3", "0", "0", "1", "-9"],
            ["F4", "S4", "0", "0", "1", "-9"],
            ["F5", "S5", "0", "0", "2", "-9"],
            ["F6", "S6", "0", "0", "1", "-9"],
            ["F7", "S7", "0", "0", "0", "-9"],
            ["F8", "S8", "0", "0", "1", "-9"],
            ["F9", "S9", "0", "0", "2", "-9"],
        ]
        with open(prefix + ".fam", "w") as o_file:
            for sample in fam_content:
                print(*sample, sep=" ", file=o_file)

        # Creating the BIM file
        bim_content = [
            ["1", "M0", "0", "123", "A", "G"],
            ["1", "M1", "0", "124", "C", "T"],
            ["2", "M2", "0", "117", "G", "C"],
        ]
        with open(prefix + ".bim", "w") as o_file:
            for marker in bim_content:
                print(*marker, sep="\t", file=o_file)

        # Creating a transposed pedfile using Plink
        out_prefix = prefix + "_transposed"
        try:
            check_call([
                self.plink_path,
                "--noweb",
                "--bfile", prefix,
                "--recode", "--transpose", "--tab",
                "--out", out_prefix,
            ], stdout=PIPE, stderr=PIPE)
        except CalledProcessError:
            self.fail("Plink could not recode file")

        # Checking the two files exists
        self.assertTrue(os.path.isfile(out_prefix + ".tped"))
        self.assertTrue(os.path.isfile(out_prefix + ".tfam"))

        # Checking the content of the TFAM file
        expected = "\n".join("\t".join(sample) for sample in fam_content)
        with open(out_prefix + ".tfam", "r") as i_file:
            self.assertEqual(expected + "\n", i_file.read())

        # Checking the content of the TPED file
        with open(out_prefix + ".tped", "r") as i_file:
            # Checking the first marker
            marker_1 = i_file.readline().rstrip("\r\n").split("\t")
            self.assertEqual(["1", "M0", "0", "123"], marker_1[:4])
            self.assertEqual(["G G", "A G", "G G", "G G", "0 0", "G G", "A G",
                              "G G", "G G", "A A"],
                             marker_1[4:])

            # Checking the second marker
            marker_2 = i_file.readline().rstrip("\r\n").split("\t")
            self.assertEqual(["1", "M1", "0", "124"], marker_2[:4])
            self.assertEqual(["C C", "T C", "C C", "C C", "C C", "C C", "C C",
                              "T C", "T T", "T C"],
                             marker_2[4:])

            # Checking the third marker
            marker_3 = i_file.readline().rstrip("\r\n").split("\t")
            self.assertEqual(["2", "M2", "0", "117"], marker_3[:4])
            self.assertEqual(["C C", "C C", "C C", "C C", "C C", "G C", "C C",
                              "C C", "C C", "C C"],
                             marker_3[4:])

            # Checking this is the end of the file
            self.assertEqual("", i_file.readline())

    @unittest.skipIf(platform.system() not in {"Darwin", "Linux", "Windows"},
                     "Plink not available for {}".format(platform.system()))
    def test_with_plink_individual_major(self):
        """Tests to read a binary file (INDIVIDUAL-major) using Plink."""
        # Checking if we need to skip
        if self.plink_path is None:
            self.skipTest(self.plink_message)

        # The genotypes
        all_genotypes = [
            [0, 1, 0, 0, -1, 0, 1, 0, 0, 2],
            [2, 1, 2, 2,  2, 2, 2, 1, 0, 1],
            [0, 0, 0, 0,  0, 1, 0, 0, 0, 0],
        ]
        transposed_genotypes = [
            [row[i] for row in all_genotypes]
            for i in range(len(all_genotypes[0]))
        ]

        # Creating the BED file (INDIVIDUAL-major)
        prefix = os.path.join(self.tmp_dir, "test_output")
        with pyplink.PyPlink(prefix, "w", "INDIVIDUAL-major") as pedfile:
            for genotypes in transposed_genotypes:
                pedfile.write_genotypes(genotypes)

        # Creating the FAM file
        fam_content = [
            ["F0", "S0", "0", "0", "1", "-9"],
            ["F1", "S1", "0", "0", "1", "-9"],
            ["F2", "S2", "0", "0", "2", "-9"],
            ["F3", "S3", "0", "0", "1", "-9"],
            ["F4", "S4", "0", "0", "1", "-9"],
            ["F5", "S5", "0", "0", "2", "-9"],
            ["F6", "S6", "0", "0", "1", "-9"],
            ["F7", "S7", "0", "0", "0", "-9"],
            ["F8", "S8", "0", "0", "1", "-9"],
            ["F9", "S9", "0", "0", "2", "-9"],
        ]
        with open(prefix + ".fam", "w") as o_file:
            for sample in fam_content:
                print(*sample, sep=" ", file=o_file)

        # Creating the BIM file
        bim_content = [
            ["1", "M0", "0", "123", "A", "G"],
            ["1", "M1", "0", "124", "C", "T"],
            ["2", "M2", "0", "117", "G", "C"],
        ]
        with open(prefix + ".bim", "w") as o_file:
            for marker in bim_content:
                print(*marker, sep="\t", file=o_file)

        # Creating a transposed pedfile using Plink
        out_prefix = prefix + "_transposed"
        try:
            check_call([
                self.plink_path,
                "--noweb",
                "--bfile", prefix,
                "--recode", "--transpose", "--tab",
                "--out", out_prefix,
            ], stdout=PIPE, stderr=PIPE)
        except CalledProcessError:
            self.fail("Plink could not recode file")

        # Checking the two files exists
        self.assertTrue(os.path.isfile(out_prefix + ".tped"))
        self.assertTrue(os.path.isfile(out_prefix + ".tfam"))

        # Checking the content of the TFAM file
        expected = "\n".join("\t".join(sample) for sample in fam_content)
        with open(out_prefix + ".tfam", "r") as i_file:
            self.assertEqual(expected + "\n", i_file.read())

        # Checking the content of the TPED file
        with open(out_prefix + ".tped", "r") as i_file:
            # Checking the first marker
            marker_1 = i_file.readline().rstrip("\r\n").split("\t")
            self.assertEqual(["1", "M0", "0", "123"], marker_1[:4])
            self.assertEqual(["G G", "A G", "G G", "G G", "0 0", "G G", "A G",
                              "G G", "G G", "A A"],
                             marker_1[4:])

            # Checking the second marker
            marker_2 = i_file.readline().rstrip("\r\n").split("\t")
            self.assertEqual(["1", "M1", "0", "124"], marker_2[:4])
            self.assertEqual(["C C", "T C", "C C", "C C", "C C", "C C", "C C",
                              "T C", "T T", "T C"],
                             marker_2[4:])

            # Checking the third marker
            marker_3 = i_file.readline().rstrip("\r\n").split("\t")
            self.assertEqual(["2", "M2", "0", "117"], marker_3[:4])
            self.assertEqual(["C C", "C C", "C C", "C C", "C C", "G C", "C C",
                              "C C", "C C", "C C"],
                             marker_3[4:])

            # Checking this is the end of the file
            self.assertEqual("", i_file.readline())

    def test_wrong_bed_format(self):
        """Tests opening a BED file with unknown format."""
        with self.assertRaises(ValueError) as cm:
            pyplink.PyPlink(self.prefix, bed_format="UNKNOWN-major")
        self.assertEqual(
            "invalid bed format: UNKNOWN-major",
            str(cm.exception),
        )

    def test_invalid_bed_format_with_r_mode(self):
        """Tests an invalid BED format with r mode."""
        with self.assertRaises(ValueError) as cm:
            pyplink.PyPlink(self.prefix, bed_format="INDIVIDUAL-major")
        self.assertEqual(
            "only SNP-major format is supported with mode 'r'",
            str(cm.exception),
        )

    def test_write_genotypes_in_r_mode(self):
        """Tests to use the 'write_genotypes' function in read mode."""
        with self.assertRaises(UnsupportedOperation) as cm:
            self.pedfile.write_genotypes([0, 0, 0])
        self.assertEqual("not available in 'r' mode", str(cm.exception))

    @mock.patch.object(pyplink, "logger")
    def test_dup_markers(self, pyplink_logger):
        """Tests when there are duplicated markers."""
        # Checking the original one has no duplicates
        self.assertEqual(len(self.pedfile.get_duplicated_markers()), 0)

        # Copying the BED and the FAM to the temporary directory
        new_prefix = os.path.join(self.tmp_dir, "with_dup")
        for suffix in (".bed", ".fam"):
            shutil.copyfile(self.prefix + suffix, new_prefix + suffix)

        # Copying the BIM file to the temporary directory
        shutil.copyfile(self.prefix + "_with_dup.bim", new_prefix + ".bim")

        # Reading the new files
        pedfile = pyplink.PyPlink(new_prefix)

        # Checking a warning was called
        self.assertTrue(pyplink_logger.warning.called)

        # Checking the BIM
        chromosomes = [1, 2, 3, 4, 4, 5, 6, 6, 6, 8]
        markers = ["rs10399749", "rs2949420:dup1", "rs2949421", "rs2691310",
                   "rs4030303:dup1", "rs4030303:dup2", "rs4030303:dup3",
                   "rs940550:dup1", "rs940550:dup2", "rs2949420:dup2"]
        positions = [45162, 45257, 45413, 46844, 72434, 72515, 77689, 78032,
                     81468, 222077]
        cms = [0, 1, 1, 2, 2, 3, 4, 4, 5, 6]
        a1s = ["G", "C", "0", "A", "0", "0", "G", "0", "G", "A"]
        a2s = ["C", "T", "0", "T", "G", "C", "A", "T", "C", "G"]

        # Getting the BIM file
        bim = pedfile.get_bim()

        # Checking the columns
        self.assertTrue(
            set(bim.columns.values) == {"chrom", "pos", "cm", "a1", "a2"}
        )

        # Checking the indexes
        self.assertTrue(set(bim.index.values) == set(markers))

        # Checking the values for the markers
        zipped = zip(markers, chromosomes, positions, cms, a1s, a2s)
        for marker, chrom, pos, cm, a1, a2 in zipped:
            self.assertEqual(chrom, bim.loc[marker, "chrom"])
            self.assertEqual(pos, bim.loc[marker, "pos"])
            self.assertEqual(cm, bim.loc[marker, "cm"])
            self.assertEqual(a1, bim.loc[marker, "a1"])
            self.assertEqual(a2, bim.loc[marker, "a2"])

        # Checking only one duplicated markers
        for i, marker in enumerate(markers):
            geno = pedfile.get_geno_marker(marker)
            np.testing.assert_array_equal(geno, self.genotypes[i])

        # Checking the list of duplicated markers
        expected_dup = dict(
            rs2949420=["rs2949420:dup1", "rs2949420:dup2"],
            rs4030303=["rs4030303:dup1", "rs4030303:dup2", "rs4030303:dup3"],
            rs940550=["rs940550:dup1", "rs940550:dup2"],
        )
        self.assertEqual(expected_dup, pedfile.get_duplicated_markers())

        # Closing the file
        pedfile.close()



#!/usr/bin/env python
"""Setup."""

# How to build source distribution
#   - python setup.py sdist --format bztar
#   - python setup.py sdist --format gztar
#   - python setup.py sdist --format zip
#   - python setup.py bdist_wheel --universal

# How to build for conda (do both with 2.7 and 3.4)
#   - cd conda_recipe
#   - conda clean -ytps; conda build purge; conda build --python $VERSION .
#   - cp $FILE ../conda_dist/linux-64
#   - conda convert -p all ../conda_dist/linux-64/$FILE -o ../conda_dist
#   - cd ../conda_dist && conda index *


import os
import sys
from setuptools import setup


MAJOR = 1
MINOR = 3
MICRO = 6
VERSION = "{}.{}.{}".format(MAJOR, MINOR, MICRO)


def write_version_file(fn=None):
    if fn is None:
        fn = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            os.path.join("pyplink", "version.py"),
        )

    content = (
        "\n# THIS FILE WAS GENERATED AUTOMATICALLY BY PYPLINK SETUP.PY\n"
        'pyplink_version = "{version}"\n'
    )

    a = open(fn, "w")
    try:
        a.write(content.format(version=VERSION))
    finally:
        a.close()


def get_requirements():
    # Initial requirements
    requirements = ["numpy >= 1.8.2", "pandas >= 0.17.1", "six >= 1.9.0"]

    # Checking if python 2 (requires mock)
    if sys.version_info[0] == 2:
        requirements.append(["mock >= 2.0.0"])

    return requirements


def setup_package():
    # Saving the version into a file
    write_version_file()

    setup(
        zip_safe=False,
        name="pyplink",
        version=VERSION,
        description="Python module to read binary Plink files.",
        author="Louis-Philippe Lemieux Perreault",
        author_email="louis-philippe.lemieux.perreault@statgen.org",
        url="https://github.com/lemieuxl/pyplink",
        license="MIT",
        packages=["pyplink", "pyplink.tests"],
        package_data={"pyplink.tests": ["data/*"], },
        test_suite="pyplink.tests.test_suite",
        install_requires=get_requirements(),
        classifiers=["Operating System :: POSIX :: Linux",
                     "Operating System :: MacOS :: MacOS X",
                     "Operating System :: Microsoft",
                     "Programming Language :: Python",
                     "Programming Language :: Python :: 2.7",
                     "Programming Language :: Python :: 3",
                     "Programming Language :: Python :: 3.7",
                     "Programming Language :: Python :: 3.8",
                     "Programming Language :: Python :: 3.9",
                     "Programming Language :: Python :: 3.10",
                     "Programming Language :: Python :: 3.11",
                     "License :: OSI Approved :: MIT License",
                     "Topic :: Scientific/Engineering :: Bio-Informatics"],
        keywords="bioinformatics format Plink binary",
    )

    return


if __name__ == "__main__":
    setup_package()
